/*
 *
 * j-chkmail - Mail Server Filter for sendmail
 *
 * Copyright (c) 2001, 2002, 2003, 2004 - Ecole des Mines de Paris
 *
 *  Auteur     : Jose Marcio Martins da Cruz
 *               Jose-Marcio.Martins@ensmp.fr
 *
 *  Historique :
 *  Creation     : janvier 2002
 *
 * This program is free software, but with restricted license :
 *
 * - j-chkmail is distributed only to registered users
 * - j-chkmail license is available only non-commercial applications,
 *   this means, you can use j-chkmail if you make no profit with it.
 * - redistribution of j-chkmail in any way : binary, source in any
 *   media, is forbidden
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * More details about j-chkmail license can be found at j-chkmail
 * web site : http://j-chkmail.ensmp.fr
 */

#ifndef __JNETCLASS_H

#define NET_UNKNOWN  0
#define NET_LOCAL    1
#define NET_DOMAIN   2
#define NET_FRIEND   4
#define NET_AUTH     8
#define NET_OTHER   16

#if OTHER_IS_KNOWN
# define NET_KNOWN   (NET_LOCAL | NET_FRIEND | NET_DOMAIN | NET_AUTH | NET_OTHER)
#else
# define NET_KNOWN   (NET_LOCAL | NET_FRIEND | NET_DOMAIN | NET_AUTH)
#endif

#define IS_FRIEND(class)  (((class) & NET_FRIEND) != 0 ? TRUE : FALSE)
#define IS_DOMAIN(class)  (((class) & NET_DOMAIN) != 0 ? TRUE : FALSE)
#define IS_LOCAL(class)   (((class) & NET_LOCAL) != 0 ? TRUE : FALSE)
#define IS_AUTH(class)    (((class) & NET_AUTH) != 0 ? TRUE : FALSE)
#define IS_OTHER(class)   (((class) & NET_OTHER) != 0 ? TRUE : FALSE)

#define IS_KNOWN(class)   (((class) & NET_KNOWN) != 0 ? TRUE : FALSE)
#define IS_UNKNOWN(class) (((class) & NET_KNOWN) == 0 ? TRUE : FALSE)

#define SET_NET_CLASS(class, which)   ((class) |= (which))

#define CLR_NET_CLASS(class, which)   ((class) &= ~(which))

#define    NET_CLASS_LABEL(class)  (IS_LOCAL(class) ? "LOCAL" : \
                                    IS_DOMAIN(class) ? "DOMAIN" : \
                                    IS_FRIEND(class) ? "FRIEND" : \
                                    IS_AUTH(class) ? "AUTH" : \
                                    IS_OTHER(class) ? "OTHER" : "UNKNOWN")

#define NET_CLASS_VALUE(label) (STRCASEEQUAL(label, "LOCAL") ? NET_LOCAL : \
				STRCASEEQUAL(label, "DOMAIN") ? NET_DOMAIN : \
				STRCASEEQUAL(label, "FRIEND") ? NET_OTHER : \
				STRCASEEQUAL(label, "OTHER") ? NET_OTHER : \
				NET_UNKNOWN)

typedef struct {
  int     class;
  char    label[32];
} netclass_T;

int                 check_host_class (char *ip, char *name, 
                                      char *value, size_t sz);

#define __JNETCLASS_H
#endif
