/*
 *
 * j-chkmail - Mail Server Filter for sendmail
 *
 * Copyright (c) 2001, 2002, 2003, 2004 - Ecole des Mines de Paris
 *
 *  Auteur     : Jose Marcio Martins da Cruz
 *               Jose-Marcio.Martins@ensmp.fr
 *
 *  Historique :
 *  Creation     : janvier 2002
 *
 * This program is free software, but with restricted license :
 *
 * - j-chkmail is distributed only to registered users
 * - j-chkmail license is available only non-commercial applications,
 *   this means, you can use j-chkmail if you make no profit with it.
 * - redistribution of j-chkmail in any way : binary, source in any
 *   media, is forbidden
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * More details about j-chkmail license can be found at j-chkmail
 * web site : http://j-chkmail.ensmp.fr
 */

#include <j-sys.h>

#include "j-chkmail.h"

#undef  DEBUG
#define DEBUG 0

#define EMAIL_INVALID    -1
#define EMAIL_FULL       0
#define EMAIL_USER       1
#define EMAIL_DOMAIN     2

static int     get_email_type(char *);
static bool    match_address(char *, char *);
static bool    match_connect_access(char *, int, char *);

/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
#define ACCESS_MSG_LEN     64

typedef struct
{
  char           connect[ACCESS_MSG_LEN];
  char           from[ACCESS_MSG_LEN];
  char           to[ACCESS_MSG_LEN];
  char           value[ACCESS_MSG_LEN];
  int            what;
}
ACCESS_REC;

static j_table_T htbl;

static pthread_mutex_t st_mutex = PTHREAD_MUTEX_INITIALIZER;

#define DATA_LOCK() \
  if (pthread_mutex_lock(&st_mutex) != 0) { \
    LOG_SYS_ERROR("pthread_mutex_lock"); \
  }

#define DATA_UNLOCK() \
  if (pthread_mutex_unlock(&st_mutex) != 0) { \
    LOG_SYS_ERROR("pthread_mutex_unlock"); \
  }

static int     access_label2val(char *);

static bool    fill_access_rec(ACCESS_REC *, char *);

static char   *next_access_token(char *);

char          *extract_email_address(char *, char *, size_t);


/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
void
dump_access_table()
{
  ACCESS_REC     p;

  printf("Let's dump j_acces_table : \n");
  if (j_table_get_first(&htbl, &p) == 0)
  {
    printf("   %-20s %-20s %-32s : %s\n", "CONNECT", "FROM", "TO", "VALUE");
    do
    {
      printf("-> %-20s %-20s %-32s : %s\n", p.connect, p.from, p.to, p.value);
    } while (j_table_get_next(&htbl, &p) == 0);
  }
}

/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
static int
access_comp(pa, pb)
     const void    *pa;
     const void    *pb;
{
  ACCESS_REC    *a = (ACCESS_REC *) pa;
  ACCESS_REC    *b = (ACCESS_REC *) pb;

  int            r;

  r = strcasecmp(a->connect, b->connect);
  if (r != 0)
    return r;
  r = strcasecmp(a->from, b->from);
  if (r != 0)
    return r;
  r = strcasecmp(a->to, b->to);
  if (r != 0)
    return r;
  r = strcasecmp(a->value, b->value);
  return r;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *****************************************************************************/
static         bool
fill_access_rec(r, s)
     ACCESS_REC    *r;
     char          *s;
{
  if ((r == NULL) || (s == NULL))
    return FALSE;

  if (strlen(s) == 0)
    return FALSE;

  if (strncasecmp(s, "to:", strlen("to:")) == 0)
  {
    s += strlen("to:");
    strlcpy(r->to, s, sizeof (r->to));
    (void) strtolower(r->to);
    return TRUE;
  }
  if (strncasecmp(s, "from:", strlen("from:")) == 0)
  {
    s += strlen("from:");
    strlcpy(r->from, s, sizeof (r->from));
    (void) strtolower(r->from);
    return TRUE;
  }
  if (strncasecmp(s, "connect:", strlen("connect:")) == 0)
  {
    s += strlen("connect:");
    strlcpy(r->connect, s, sizeof (r->connect));
    return TRUE;
  }

  strlcpy(r->from, s, sizeof (r->from));
  strlcpy(r->to, s, sizeof (r->to));

  return TRUE;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *****************************************************************************/
static char   *
next_access_token(s)
     char          *s;
{
  char          *p = NULL;

  if (s == NULL)
    return NULL;

  if (((p = strchr(s, '/')) != NULL))
  {
    *p = '\0';
    p++;
    if (*p != '\0')
      return p;
  }
  return NULL;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *****************************************************************************/
static int
add_access_rec(vk, vv)
     void          *vk;
     void          *vv;
{
  char          *k = (char *) vk;
  char          *v = (char *) vv;
  ACCESS_REC     r;

  char          *ka = NULL, *kb = NULL, *kc = NULL;

  if (k == NULL)
    return 0;

  memset(&r, 0, sizeof (r));
  if (k == NULL)
    k = "";
  if (v == NULL)
    v = "";

  ka = k;
  kb = next_access_token(ka);
  kc = next_access_token(kb);

  (void) fill_access_rec(&r, ka);
  (void) fill_access_rec(&r, kb);
  (void) fill_access_rec(&r, kc);

  if ((strlen(r.from) == 0) && (strlen(r.to) == 0) && (strlen(r.connect) == 0))
  {
    LOG_MSG_WARNING("No To and No From...");
    return 0;
  }

  strlcpy(r.value, v, sizeof (r.value));

  r.what = access_label2val(v);

  return j_table_add(&htbl, &r);
}

/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
static         bool
read_it(path, tag)
     char          *path;
     char          *tag;
{
  int            r;

  r = j_rd_text_file(path, RD_TWO_COLUMN, RD_NORMAL, tag, add_access_rec);

  return r >= 0;
}

bool
load_access_table(cfdir, fname)
     char          *cfdir;
     char          *fname;
{
  int            res = 0;
  static int     htbl_ok = FALSE;
  bool           result = FALSE;

  DATA_LOCK();
  if (htbl_ok == FALSE)
  {
    memset(&htbl, 0, sizeof (htbl));
    res = j_table_init(&htbl, sizeof (ACCESS_REC), 256, access_comp);
    if (res == 0)
      htbl_ok = TRUE;
  }

  if (res == 0)
    res = j_table_clear(&htbl);

  if (res == 0)
    result = read_conf_data_file(cfdir, fname, "j-access", read_it);

  DATA_UNLOCK();

  return result;
}

/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
#define DLEVEL        15

bool
check_access(connect, ip_class, hostname, from, to, msg)
     char          *connect;
     int            ip_class;
     char          *hostname;
     char          *from;
     char          *to;
     char          *msg;
{
  ACCESS_REC    *q;
  bool           res = FALSE;
  char          *pfrom = NULL;
  char          *pto = NULL;

  /* XXX for the while, only From address is checked */
#if 0
  if (from == NULL)
    return res;
#endif

  if (from != NULL)
  {
    if ((pfrom = strdup(from)) == NULL)
    {
      LOG_SYS_ERROR("strdup(from) error");
      return res;
    }
  }

  if (pfrom != NULL)
    (void) extract_email_address(pfrom, from, strlen(pfrom) + 1);

  if (to != NULL)
  {
    if ((pto = strdup(to)) == NULL)
    {
      LOG_SYS_ERROR("strdup(to) error");
      FREE(pfrom);
      return res;
    }
  }

  if (pto != NULL)
    (void) extract_email_address(pto, to, strlen(pto) + 1);

  DATA_LOCK();

  q = (ACCESS_REC *) j_table_get_first_ptr(&htbl);
  for (; q != NULL; q = (ACCESS_REC *) j_table_get_next_ptr(&htbl))
  {
    MESSAGE_INFO(DLEVEL, "Trying   : (%s) (%s) (%s) (%s) (%s) : %s",
                 q->connect,
                 q->from,
                 STRNULL(pfrom, ""),
                 STRNULL(to, ""), q->value, STRBOOL(res, "YES", "NO"));

    if ((msg != NULL) && (strcasecmp(msg, q->value) != 0))
      continue;
    MESSAGE_INFO(DLEVEL, "  MSG     passed");

    if ((pfrom != NULL) && !match_address(pfrom, q->from))
      continue;
    MESSAGE_INFO(DLEVEL, "  SENDER  passed");

    if (!match_connect_access(connect, ip_class, q->connect))
      continue;
    MESSAGE_INFO(DLEVEL, "  CONNECT passed");

    if ((pto != NULL) && !match_address(pto, q->to))
      continue;
    MESSAGE_INFO(DLEVEL, "  RCPT    passed");

    res = TRUE;

    MESSAGE_INFO(DLEVEL, "Gotcha   : (%s) (%s) (%s) (%s) (%s) : %s",
                 q->connect,
                 q->from,
                 STRNULL(pfrom, "FROM"),
                 STRNULL(pto, "TO"), q->value, STRBOOL(res, "YES", "NO"));


    break;
  }

  DATA_UNLOCK();

  MESSAGE_INFO(DLEVEL, "Match    : %s %s %s : %s %d",
               STRNULL(connect, "-"), STRNULL(from, "-"),
               STRNULL(msg, "-"), STRBOOL(res, "YES", "NO"), res);

  FREE(pfrom);
  FREE(pto);

  return res;
}


/* ****************************************************************************
 *                                                                            * 
 *                                                                            *
 **************************************************************************** */
static int
access_label2val(code)
     char          *code;
{
  if (code == NULL)
    return ACCESS_DEFAULT;

  if (strcasecmp(code, "OK") == 0)
    return ACCESS_OK;

  if (strcasecmp(code, "REJECT") == 0)
    return ACCESS_REJECT;

  return ACCESS_DEFAULT;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *****************************************************************************/
static int
get_email_type(email)
     char          *email;
{
  int            n;

  if ((email == NULL) || (strlen(email) == 0))
    return EMAIL_INVALID;

  n = strcspn(email, "@");

  if ((n == 0) || (n == strlen(email)))
    return EMAIL_DOMAIN;

  if (n == (strlen(email) - 1))
    return EMAIL_USER;

  return EMAIL_FULL;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *****************************************************************************/
static         bool
match_connect_access(connect, class, pattern)
     char          *connect;
     int            class;
     char          *pattern;
{
  if (pattern == NULL)
    return FALSE;

  if ((strlen(pattern) == 0) || (strcmp(pattern, "*") == 0))
    return TRUE;

  if ((strcasecmp("KNOWN", pattern) == 0) && IS_KNOWN(class))
    return TRUE;

  if ((strcasecmp("UNKNOWN", pattern) == 0) && IS_UNKNOWN(class))
    return TRUE;

  if ((strcasecmp("LOCAL", pattern) == 0) && IS_LOCAL(class))
    return TRUE;

  if ((strcasecmp("DOMAIN", pattern) == 0) && IS_DOMAIN(class))
    return TRUE;

  if ((strcasecmp("FRIEND", pattern) == 0) && IS_FRIEND(class))
    return TRUE;

  if ((strcasecmp("AUTH", pattern) == 0) && IS_AUTH(class))
    return TRUE;

  if (connect == NULL)
    return FALSE;

  if (strlen(connect) > 0)
  {
    int            n = strlen(pattern);
    int            m = strlen(connect);

    if (n > m)
      return FALSE;

    if (n == m)
    {
      if (strcasecmp(connect, pattern) != 0)
        return FALSE;
    } else
    {                           /* n < m */
      if (strncasecmp(connect, pattern, n) != 0)
        return FALSE;
      if (connect[n] != '.')
        return FALSE;
    }
    return TRUE;
  }
  return FALSE;
}

/*****************************************************************************
 *                                                                           *
 *                                                                           *
 *****************************************************************************/
static         bool
match_address(email, pattern)
     char          *email;
     char          *pattern;
{
  int            email_type = EMAIL_INVALID;
  int            n;
  char          *p;

  if ((pattern == NULL) || (strlen(pattern) == 0))
    return TRUE;

  if ((email == NULL) || (strlen(email) == 0))
    return FALSE;

#if _FFR_ACCESS_REGEX
  return strexpr(email, pattern, NULL, NULL, TRUE);
#else
  email_type = get_email_type(pattern);

  switch (email_type)
  {
    case EMAIL_INVALID:
      return FALSE;
    case EMAIL_FULL:
      if (strcasecmp(email, pattern) == 0)
        return TRUE;
      break;
    case EMAIL_USER:
      n = strlen(email);
      if (strncasecmp(email, pattern, n) == 0)
        return TRUE;
      break;
    case EMAIL_DOMAIN:
      p = strstr(email, pattern);
      if (p == NULL)
        return FALSE;
      if (strcasecmp(p, pattern) == 0)
        return TRUE;
      break;
    default:
      break;
  }
  return FALSE;
#endif
}
