/*
 *
 * j-chkmail - Mail Server Filter for sendmail
 *
 * Copyright (c) 2001, 2002, 2003, 2004 - Ecole des Mines de Paris
 *
 *  Auteur     : Jose Marcio Martins da Cruz
 *               Jose-Marcio.Martins@ensmp.fr
 *
 *  Historique :
 *  Creation     : janvier 2002
 *
 * This program is free software, but with restricted license :
 *
 * - j-chkmail is distributed only to registered users
 * - j-chkmail license is available only non-commercial applications,
 *   this means, you can use j-chkmail if you make no profit with it.
 * - redistribution of j-chkmail in any way : binary, source in any
 *   media, is forbidden
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * More details about j-chkmail license can be found at j-chkmail
 * web site : http://j-chkmail.ensmp.fr
 */

#include <j-sys.h>

#include "j-chkmail.h"

/*****************************************************************************
 *                                                                           * 
 *                                                                           *
 *****************************************************************************/
#define SET_LABEL(s, sz, label)			\
  do {						\
    if ((s) != NULL && (sz) > 0)		\
      strlcpy((s), (label), (sz));		\
  } while (0)

int
check_host_class(ip, name, label, sz)
     char               *ip;
     char               *name;
     char               *label;
     size_t              sz;
{
  int                 ret = NET_UNKNOWN;
  char                buf[256];
  bool                found = FALSE;

  SET_LABEL(label, sz, "UNKNOWN");

  if (ip == NULL)
    return NET_UNKNOWN;

  if (STREQUAL(ip, "127.0.0.1")) {
    SET_LABEL(label, sz, "LOCAL");
    return NET_LOCAL;
  }

  if (STRNCASEEQUAL(ip, "ipv6:", strlen("ipv6")))
    ip += strlen("ipv6:");

#if 0
  if (STREQUAL(ip, "::1")) {
    SET_LABEL(label, sz, "OTHER");
    return NET_OTHER;
  }
#endif

  memset(buf, 0, sizeof (buf));

  found = db_policy_check("NetClass", ip, buf, sizeof (buf));
  if (!found && name != NULL)
    found = db_policy_check("NetClass", name, buf, sizeof (buf));

  if (found)
  {
    MESSAGE_INFO(12, " NetClass : %s %s %s", ip, STRNULL(name, "-"), buf);
    ret = NET_OTHER;
    SET_LABEL(label, sz, buf);

    if (strcasecmp(buf, "LOCAL") == 0)
      ret = NET_LOCAL;

    if (strcasecmp(buf, "DOMAIN") == 0)
      ret = NET_DOMAIN;

    if (strcasecmp(buf, "FRIEND") == 0)
      ret = NET_FRIEND;

    if (strcasecmp(buf, "UNKNOWN") == 0)
      ret = NET_UNKNOWN;
  }

  return ret;
}
