/*
 *
 * j-chkmail - Mail Server Filter for sendmail
 *
 * Copyright (c) 2001, 2002, 2003, 2004 - Ecole des Mines de Paris
 *
 *  Auteur     : Jose Marcio Martins da Cruz
 *               Jose-Marcio.Martins@ensmp.fr
 *
 *  Historique :
 *  Creation     : janvier 2002
 *
 * This program is free software, but with restricted license :
 *
 * - j-chkmail is distributed only to registered users
 * - j-chkmail license is available only non-commercial applications,
 *   this means, you can use j-chkmail if you make no profit with it.
 * - redistribution of j-chkmail in any way : binary, source in any
 *   media, is forbidden
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * More details about j-chkmail license can be found at j-chkmail
 * web site : http://j-chkmail.ensmp.fr
 */

#include <j-sys.h>

#include "j-libjc.h"

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
LIST_T        *
linked_list_add(head, s, nb, data, size)
     LIST_T        *head;
     char          *s;
     int            nb;
     void          *data;
     size_t         size;
{
  LIST_T        *p;

  if ((s == NULL) || (strlen(s) == 0))
    return NULL;

  for (p = head; p != NULL; p = p->next)
  {
    if (strcasecmp(p->s, s) == 0)
      break;
  }

  if (p != NULL)
  {
    p->n += nb;
    return head;
  }

  p = (LIST_T *) malloc(sizeof (LIST_T));
  if (p == NULL)
  {
    LOG_SYS_ERROR("malloc(LIST_T)");
    return NULL;
  }
  memset(p, 0, sizeof (LIST_T));
  p->next = head;

  if ((p->s = strdup(s)) == NULL)
  {
    LOG_SYS_ERROR("strdup(LIST_T)");
    FREE(p);
    return NULL;
  }
  if (data != NULL && size > 0)
  {
    if ((p->data = malloc(size)) != NULL)
    {
      memcpy(p->data, data, size);
      p->size = size;
    } else
      LOG_SYS_ERROR("strdup(LIST_T)");
  }
  p->n = nb;
  head = p;

  return head;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
LIST_T        *
linked_list_find(head, s)
     LIST_T        *head;
     char          *s;
{
  LIST_T        *p;

  if ((s == NULL) || (strlen(s) == 0))
    return NULL;

  for (p = head; p != NULL; p = p->next)
  {
    if (strcasecmp(p->s, s) == 0)
      return p;
  }

  return NULL;
}


/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
LIST_T        *
linked_list_clear(head, func)
     LIST_T        *head;
     LISTCLEAR_F    func;
{
  LIST_T        *p;

  while (head != NULL)
  {
    p = head;
    head = p->next;
    FREE(p->s);
    if (func != NULL && p->data != NULL)
      func(p->data);
    FREE(p->data);
    FREE(p);
  }

  return head;
}
