/*
 *
 * j-chkmail - Mail Server Filter for sendmail
 *
 * Copyright (c) 2001-2009 - Ecole des Mines de Paris
 *
 *  Auteur       : Jose Marcio Martins da Cruz
 *                 Jose-Marcio.Martins@ensmp.fr
 *
 *  Historique   :
 *  Creation     : Sat Dec 19 23:42:11 CET 2009
 *
 * This program is free software, but with restricted license :
 *
 * - j-chkmail is distributed only to registered users
 * - j-chkmail license is available only non-commercial applications,
 *   this means, you can use j-chkmail if you make no profit with it.
 * - redistribution of j-chkmail in any way : binary, source in any
 *   media, is forbidden
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * More details about j-chkmail license can be found at j-chkmail
 * web site : http://j-chkmail.ensmp.fr
 */


#include <j-sys.h>
#include <j-chkmail.h>
#include <j-callback.h>

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */

static kstats_T     callback_st[CALLBACK_LAST + 1];
static kstats_T     callback_gst;

static pthread_mutex_t mutex = PTHREAD_MUTEX_INITIALIZER;

static bool         st_ok = FALSE;


/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static void
stats_init()
{
  int                 i;

  MUTEX_LOCK(&mutex);

  for (i = CALLBACK_FIRST; i <= CALLBACK_LAST; i++)
    kstats_reset(&callback_st[i]);
  kstats_reset(&callback_gst);
  st_ok = TRUE;

  MUTEX_UNLOCK(&mutex);
}


/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
bool
callback_stats_update(callback, dt)
     int                 callback;
     timems_T            dt;
{
  if (callback < CALLBACK_FIRST || callback > CALLBACK_LAST)
    return FALSE;

  if (!st_ok)
    stats_init();

  MUTEX_LOCK(&mutex);
  kstats_update(&(callback_st[callback]), (double) dt);
  kstats_update(&callback_gst, (double) dt);
  MUTEX_UNLOCK(&mutex);

  return TRUE;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
bool
callback_stats_dump(fd)
     int                 fd;
{
  int                 i;

  MUTEX_LOCK(&mutex);
  if (fd >= 0)
  {
    FD_PRINTF(fd, "   Callback Handling Times Statistics\n");
    FD_PRINTF(fd, "*  %-15s : %8s %10s %10s %10s %10s\n",
              "Callback", "Count", "Minimum", "Maximum", "Mean", "Std. Dev.");
    FD_PRINTF(fd, "*  %-15s : %8s %10s %10s %10s %10s\n",
              "********", "********", "*********", "*********",
              "*********", "*********");
  } else
  {
    MESSAGE_INFO(9, "   Callback Handling Times Statistics");
    MESSAGE_INFO(9, "*  %-15s : %8s %10s %10s %10s %10s",
                 "Callback", "Count", "Minimum", "Maximum", "Mean",
                 "Std. Dev.");
    MESSAGE_INFO(9, "*  %-15s : %8s %10s %10s %10s %10s", "********",
                 "********", "*********", "*********", "*********",
                 "*********");
  }
  for (i = CALLBACK_FIRST; i <= CALLBACK_LAST; i++)
  {
    if (fd >= 0)
      FD_PRINTF(fd, "*  %-15s : %8d %10.1f %10.1f %10.1f %10.1f  ms\n",
                CALLBACK_LABEL(i),
                kcount(&callback_st[i]),
                kmin(&callback_st[i]),
                kmax(&callback_st[i]),
                kmean(&callback_st[i]), kstddev(&callback_st[i]));
    else
      MESSAGE_INFO(9, "*  %-15s : %8d %10.1f %10.1f %10.1f %10.1f  ms",
                   CALLBACK_LABEL(i),
                   kcount(&callback_st[i]),
                   kmin(&callback_st[i]),
                   kmax(&callback_st[i]),
                   kmean(&callback_st[i]), kstddev(&callback_st[i]));
  }
  if (fd >= 0)
  {
    FD_PRINTF(fd, "*  %-15s : %8s %10s %10s %10s %10s\n",
              "********", "********", "*********", "*********",
              "*********", "*********");
    FD_PRINTF(fd, "*  %-15s : %8d %10.1f %10.1f %10.1f %10.1f  ms\n",
              "Global",
              kcount(&callback_gst),
              kmin(&callback_gst),
              kmax(&callback_gst),
              kmean(&callback_gst), kstddev(&callback_gst));
  } else
  {
    MESSAGE_INFO(9, "*  %-15s : %8s %10s %10s %10s %10s",
                 "********", "********", "*********", "*********",
                 "*********", "*********");
    MESSAGE_INFO(9, "*  %-15s : %8d %10.1f %10.1f %10.1f %10.1f  ms",
                 "Global",
                 kcount(&callback_gst),
                 kmin(&callback_gst),
                 kmax(&callback_gst),
                 kmean(&callback_gst), kstddev(&callback_gst));
  }

  MUTEX_UNLOCK(&mutex);

  return TRUE;
}
