/*
 *
 * j-chkmail - Mail Server Filter for sendmail
 *
 * Copyright (c) 2001, 2002, 2003, 2004 - Ecole des Mines de Paris
 *
 *  Auteur     : Jose Marcio Martins da Cruz
 *               Jose-Marcio.Martins@ensmp.fr
 *
 *  Historique :
 *  Creation     : janvier 2002
 *
 * This program is free software, but with restricted license :
 *
 * - j-chkmail is distributed only to registered users
 * - j-chkmail license is available only non-commercial applications,
 *   this means, you can use j-chkmail if you make no profit with it.
 * - redistribution of j-chkmail in any way : binary, source in any
 *   media, is forbidden
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * More details about j-chkmail license can be found at j-chkmail
 * web site : http://j-chkmail.ensmp.fr
 */

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static sfsistat     eom_check_xfiles(SMFICTX * ctx, attachment_T * ahead,
                                     bool * done);
static sfsistat     eom_check_virus(SMFICTX * ctx, attachment_T * ahead,
                                    bool * done);
static sfsistat     eom_check_content(SMFICTX * ctx, bool * done);

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
sfsistat
mlfi_eom(ctx)
     SMFICTX            *ctx;

{
  CTXPRIV_T          *priv = MLFIPRIV(ctx);
  char                hbuf[1024];
  int                 result = SMFIS_CONTINUE;
  attachment_T       *ahead = NULL;
  int                 nb_files = 0;
  int                 nb_exefiles = 0;
  char               *log_msg = NULL;
  char                log_msg_buf[1024];

  time_t              now = time(NULL);

  INIT_CALLBACK_DELAY();
  stats_inc(STAT_MSGS, 1);
  if (priv == NULL)
    return SMFIS_TEMPFAIL;

  INIT_CALLBACK(priv, CALLBACK_EOM);

  sm_macro_update(ctx, priv->sm);

  if (!spool_file_close(priv))
    LOG_MSG_WARNING("spool_file_close error");

  if (priv->rej_greyrcpt > 0)
  {
    priv->rej_greymsgs++;
    stats_inc(STAT_GREY_MSGS, 1);
  }

  priv->nb_msgs++;
  (void) smtprate_add_entry(RATE_VOLUME, priv->peer_addr, priv->peer_name,
                            (priv->msg_size + 512) / 1024, now);
  (void) smtprate_add_entry(RATE_MSGS, priv->peer_addr, priv->peer_name, 1,
                            now);

  /*
   ** Checking Delivery Notification Requests
   */
  if (IS_UNKNOWN(priv->netclass.class))
  {
    bool                do_it;

    do_it = (cf_get_int(CF_DROP_DELIVERY_NOTIFICATION_REQUEST) == OPT_YES);

    char              **hdr = NULL;
    char               *unwanted_hdrs[] = {
      "Disposition-Notification-To",
      "X-Confirm-Reading-To",
      "Generate-Delivery-Report",
      "Disclose-Recipients",
      NULL
    };

    for (hdr = unwanted_hdrs; *hdr != NULL; hdr++)
    {
      int                 nb = 0;
      header_T           *h = priv->headers;

      h = priv->headers;
      while ((h = get_msgheader_next(h, *hdr)) != NULL)
      {
        if (h->value == NULL)
          continue;

        nb++;
        log_msg_context(ctx, *hdr);
        MESSAGE_INFO(10, "%s SPAM CHECK - %s : %s",
                     CONNID_STR(priv->id), *hdr, h->value);
      }
      if (do_it)
      {
        while (nb > 0)
          smfi_chgheader(ctx, *hdr, nb--, NULL);
      }
    }
  }

  if (cf_get_int(CF_PRESENCE) == OPT_SHOW)
  {
    char               *p = NULL;
    char               *url;
    char                host[512];

    memset(host, 0, sizeof (host));
    if (!get_hostname(host, sizeof (host)))
      snprintf(host, sizeof (host), "localhost");
    /* add a header to the message announcing our presence */
    switch (cf_get_int(CF_J_HOSTNAME))
    {
      case OPT_SYSTEM:
        break;
      case OPT_SENDMAIL:
        p = smfi_getsymval(ctx, "j");
        break;
      default:
        p = cf_get_str(CF_J_HOSTNAME);
        break;
    }

    if ((p != NULL) && (strlen(p) > 0))
      strlcpy(host, p, sizeof (host));
    url = cf_get_str(CF_FILTER_URL);
    url = STRNULL(url, "");
    if (strcasecmp(url, "http://j-chkmail.ensmp.fr") == 0)
      url = "http : // j-chkmail dot ensmp dot fr";
    if (strcasecmp(url, "NO") == 0)
      url = "";
    if (strlen(url) > 0)
    {
      snprintf(hbuf, sizeof hbuf,
               "at %s with ID %s by Joe's j-chkmail (%s)!",
               host, CONNID_STR(priv->id), url);
    } else
    {
      snprintf(hbuf, sizeof hbuf,
               "at %s with ID %s by Joe's j-chkmail !", host,
               CONNID_STR(priv->id));
    }
    smfi_addheader(ctx, "X-Miltered", hbuf);
    snprintf(hbuf, sizeof (hbuf), "%s/%s/%s/%s/%s",
             CONNID_STR(priv->id),
             priv->peer_addr, priv->peer_name, priv->helohost, priv->env_from);
    smfi_addheader(ctx, "X-j-chkmail-Enveloppe", hbuf);
  }

  /*
   ** remove old headers 
   **/
  {
    int                 argc;
    char               *argv[32];
    char                buf[512], *p = NULL;
    int                 i, n;

    p = cf_get_str(CF_REMOVE_HEADERS);
    if (p != NULL)
      strlcpy(buf, p, sizeof (buf));
    else
      memset(buf, 0, sizeof (buf));
    argc = str2tokens(buf, 32, argv, ", ");
    for (i = 0; i < argc; i++)
    {
      if (strlen(argv[i]) == 0 || STRCASEEQUAL(argv[i], "none"))
        break;

      n = count_msgheader_attr(priv->headers, argv[i]);
      while (n > 0)
      {
        (void) smfi_chgheader(ctx, argv[i], n, NULL);
        n--;
      }
    }
  }

  /* 
   ** remove old scores
   */
  {
    int                 argc;
    char               *argv[32];
    char                buf[512], *p = NULL;
    int                 n;

    p = cf_get_str(CF_REMOVE_SCORES);
    if (p != NULL)
      strlcpy(buf, p, sizeof (buf));
    else
      memset(buf, 0, sizeof (buf));
    argc = str2tokens(buf, 32, argv, ", ");

    if (argc > 0)
    {
      n = count_msgheader_attr(priv->headers, "x-j-chkmail-score");
      while (n > 0)
      {
        bool                do_it = FALSE;

        header_T           *h;
        int                 i;

        h = get_msgheader_index_2(priv->headers, "X-j-chkmail-Score", n);
        if (h == NULL)
          break;
        if (h->value == NULL || strlen(h->value) == 0)
          continue;

        for (i = 0; i < argc; i++)
        {
          char               *arg = argv[i];
          bool                preserve = FALSE;
          char                expr[256];

          if (strlen(arg) == 0)
          {
            do_it = FALSE;
            break;
          }
          if (STRCASEEQUAL(arg, "none"))
          {
            do_it = FALSE;
            break;
          }
          if (STRCASEEQUAL(arg, "all"))
          {
            do_it = TRUE;
            break;
          }

          if (*arg == '!')
          {
            preserve = TRUE;
            arg++;
          }
          if (strlen(arg) == 0)
          {
            do_it = FALSE;
            break;
          }


          /* "MSGID : %s on %s : j-chkmail score : %s : %d/%d %d %5.3f -> %d", */
          snprintf(expr, sizeof (expr), "on [^ ]*%s : j-chkmail score", arg);
          if (strexpr(h->value, expr, NULL, NULL, TRUE))
          {
            if (!preserve)
              do_it = TRUE;
            break;
          }
        }

        if (do_it)
          (void) smfi_chgheader(ctx, "X-j-chkmail-Score", n, NULL);

        n--;
      }
    }
  }

  /*
   ** Let's finish attached files name extraction
   **
   */
  if (priv->tcontent.field_type != CT_NONE)
    save_content_field(&priv->tcontent, &priv->lcontent);
  if ((result == SMFIS_CONTINUE) && (priv->lcontent != NULL))
  {
    nb_files = extract_attachments(priv->lcontent, &ahead);
    priv->nb_files += nb_files;
    stats_inc(STAT_FILES, nb_files);
  }

  /*
   ** No message check before this point.
   */
#if 0
  /* Message with passport ??? */
  if (priv->nb_rcpt == 1 && priv->pass_ok)
    goto fin;
#endif

  /*
   ** update bad recipients
   */
  if (result == SMFIS_CONTINUE)
  {
    int                 n;

    if ((n = update_nb_badrcpts(ctx)) > 0)
      MESSAGE_INFO(19, "%s : BAD RECIPIENTS : %d", CONNID_STR(priv->id), n);
    priv->nb_mbadrcpt = n;
    MESSAGE_INFO(19, "mlfi_eom : bad = %d", n);
    result = check_nb_badrcpts(ctx);
    if (result != SMFIS_CONTINUE)
      priv->rej_badrcpt = TRUE;
  }
  if (result != SMFIS_CONTINUE)
    goto fin;

  /*
   ** Check for bad NULL SENDER : null sender with more than one recipient
   */
  if ((result == SMFIS_CONTINUE) && IS_UNKNOWN(priv->netclass.class))
  {
    if (cf_get_int(CF_REJECT_BAD_NULL_SENDER) != OPT_OK)
    {
      if ((priv->env_nb_rcpt > 1) && (strstr(priv->env_from, "<>") != NULL))
      {
        result = SMFIS_REJECT;
        (void) jsmfi_setreply(ctx, "550", "5.7.1", MSG_BAD_NULL_SENDER);
        log_msg_context(ctx, MSG_BAD_NULL_SENDER);
      }
    }
  }
  if (result != SMFIS_CONTINUE)
    goto fin;

  /* 
   ** Check for empty messages or too short messages 
   */
  if ((priv->body_nb == 0 || priv->msg_short)
      && IS_UNKNOWN(priv->netclass.class))
  {
    char              **cmd;
    header_T           *h = priv->headers;

    priv->msg_short = TRUE;
    h = priv->headers;
    while ((h = get_msgheader_next(h, "Subject")) != NULL)
    {
      if (h->value == NULL)
        continue;
      for (cmd = SYMPA_CMDS; (*cmd != NULL); cmd++)
      {
        MESSAGE_INFO(12, "Checking Subject %s %s", h->value, *cmd);
        if (strexpr(h->value, *cmd, NULL, NULL, TRUE))
        {
          priv->msg_short = FALSE;
          break;
        }
      }
    }
  }

  if (priv->msg_short && IS_UNKNOWN(priv->netclass.class))
  {
    int                 nb;

    nb = livehistory_add_entry(priv->peer_addr, time(NULL), 1, LH_EMPTYMSGS);
    MESSAGE_INFO(9, "%s Empty or too short message : %d bytes",
                 CONNID_STR(priv->id), priv->msg_size);
    if (cf_get_int(CF_REJECT_SHORT_BODIES) == OPT_YES)
    {
      bool                doit = FALSE;

      doit =
        check_policy_all_rcpts("RejectShortMsgs", priv->peer_addr,
                               priv->peer_name, CTX_NETCLASS_LABEL(priv),
                               priv->env_from, priv->env_rcpt, TRUE,
                               OPT_DEFAULT);
      if (doit)
      {
        result = SMFIS_REJECT;
        (void) jsmfi_setreply(ctx, "550", "5.7.1", MSG_EMPTY_MSG);
        log_msg_context(ctx, MSG_EMPTY_MSG);
      }
    }

    if (nb > 0);
  }

  if (result != SMFIS_CONTINUE)
    goto fin;

#if 0
  /*
   ** Check DomainKeys
   */
  dkresult = check_domainkeys(ctx);
  if (dkresult...)
  {

  }

  if (result != SMFIS_CONTINUE)
    goto fin;

#endif

#if _FFR_MODULES
  /* 
   ** j-chkmail modules
   **
   */
  if (do_module_callback(ctx, 0, &result))
    goto fin;
  if (result != SMFIS_CONTINUE)
    goto fin;
#endif             /* _FFR_MODULES */


#if 1
  /*
   **
   */
#define DEFAULT_ORDER     "XFILES,VIRUS,CONTENT"

  {
    char                buf[256];
    char               *argv[32];
    int                 argc;
    int                 i;

    char               *env = getenv("ORDER_CHECKS");

    if (env == NULL || strlen(env) == 0)
      env = DEFAULT_ORDER;
    strlcpy(buf, env, sizeof (buf));

    argc = str2tokens(buf, 32, argv, " ,");
    for (i = 0; i < argc; i++)
    {
      bool                done = FALSE;

      if (STRCASEEQUAL(argv[i], "xfiles"))
        result = eom_check_xfiles(ctx, ahead, &done);
      if (STRCASEEQUAL(argv[i], "virus"))
        result = eom_check_virus(ctx, ahead, &done);
      if (STRCASEEQUAL(argv[i], "content"))
        result = eom_check_content(ctx, &done);
      if (done || result != SMFIS_CONTINUE)
        break;
    }
  }

#endif

  if (result != SMFIS_CONTINUE)
    goto fin;

  if (cf_get_int(CF_LOG_ATTACHMENTS) == OPT_YES)
    log_attached_files(CONNID_STR(priv->id), priv->peer_addr, ahead);

  {
    int                 n;

    if ((n = update_nb_badrcpts(ctx)) > 0)
      MESSAGE_INFO(19, "%s : BAD RECIPIENTS : %d", CONNID_STR(priv->id), n);
    priv->nb_cbadrcpt += n;
    priv->nb_mbadrcpt = 0;
    MESSAGE_INFO(19, "mlfi_abort : bad = %d", n);
  }

fin:
  /*
   **
   ** Let's log what we found ...
   **
   */
  if (priv->save_msg)
    log_quarantine(ctx, ahead);

  if ((cf_get_int(CF_LOG_ATTACHMENTS) == OPT_YES) || (log_msg != NULL))
  {
    log_msg = STRNULL(log_msg, "*** ATTACHMENTS : ");
    log_msg_context(ctx, log_msg);
#if 0
    if (nb_files > 0)
    {
      attachment_T       *p;

      p = ahead;
      while (p != NULL)
      {
        char               *serror;

        serror = STRBOOL(p->exefile, "SUSPECT ???", "CLEAN");
        MESSAGE_INFO(10, "%s : **** (%-11s) : %-10s %-30s %s\n",
                     CONNID_STR(priv->id),
                     serror, STRNULL(p->disposition, ""),
                     STRNULL(p->mimetype, ""), STRNULL(p->name, ""));
        p = p->next;
      }
    }
#endif
  }

  /*
   **
   ** Everything is done ! Let's clean up the house !
   **
   */
  if (ahead != NULL)
    free_attachment_list(ahead);
  ahead = NULL;
  (void) mlfi_cleanup(ctx, FALSE);
  CHECK_CALLBACK_DELAY();
  return result;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
eom_check_xfiles(ctx, ahead, done)
     SMFICTX            *ctx;
     attachment_T       *ahead;
     bool               *done;
{
  CTXPRIV_T          *priv = MLFIPRIV(ctx);
  sfsistat            result = SMFIS_CONTINUE;
  time_t              now = 0;

  if (done != NULL)
    *done = FALSE;
  if ((cf_get_int(CF_XFILES) == OPT_OK) || !shall_check_xfiles(ctx))
    return result;

  now = time(NULL);
  /*
   **
   ** Let's check attached files
   **
   */
  {
    attachment_T       *p = NULL;
    char                str_action[64];
    int                 nb_exefiles = 0;
    int                 nb_files = 0;
    char               *log_msg = NULL;

    memset(str_action, 0, sizeof (str_action));
    p = ahead;
    while (p != NULL)
    {
      memset(str_action, 0, sizeof (str_action));
      if (!p->exefile)
        p->exefile = check_xfiles(p->name, p->mimetype, priv->msg_size,
                                  str_action, sizeof (str_action));
      if (p->exefile
          && (strlen(str_action) == 0 || !STRCASEEQUAL(str_action, "ok")))
        nb_exefiles++;
      p = p->next;
      nb_files++;
    }

    if (nb_files > 0)
    {
      priv->nb_xfiles += nb_exefiles;
      stats_inc(STAT_XFILES, nb_exefiles);
    }

    /* 
     **
     ** Is there any XFILE ?
     **
     */

    if (nb_exefiles > 0)
    {
      /* remplacer corps du message */
      /* message de log */
      /* ajoute l'expediteur dans la liste des destinataires */

      int                 res_action = cf_get_int(CF_XFILES);
      smtp_reply_T        reply;

      if (0)
      {
        char               *xstatus = NULL;

        xstatus = cf_get_str(CF_XSTATUS_HEADER);
        if (xstatus == NULL || strlen(xstatus) == 0)
          xstatus = "X-j-chkmail-Status";
        smfi_addheader(ctx, xstatus, "Spam HI");
      }

      if (cf_get_int(CF_XFILE_SAVE_MSG) == OPT_YES)
        DO_QUARANTINE_MESSAGE(priv, WHY_XFILE, NULL);

      log_msg = MSG_SHORT_XFILE;
      memset(&reply, 0, sizeof (reply));
      jc_fill_reply(&reply, "550", "5.6.0", MSG_XFILE, SMFIS_REJECT);
      if (strlen(str_action) > 0)
      {
        if (STRCASEEQUAL(str_action, "ok"))
        {
          res_action = OPT_OK;
          goto fin;
        }
        if (STRCASEEQUAL(str_action, "reject"))
        {
          res_action = OPT_REJECT;
          goto do_action;
        }
        if (STRCASEEQUAL(str_action, "notify"))
        {
          res_action = OPT_NOTIFY;
          goto do_action;
        }
        if (STRCASEEQUAL(str_action, "discard"))
        {
          res_action = OPT_DISCARD;
          goto do_action;
        }
        if (STRCASEEQUAL(str_action, "x-header"))
        {
          res_action = OPT_X_HEADER;
          goto do_action;
        }
#if 0
        {
          char                buf[256];

          if (query_policy
              ("XfilesAction", str_action, buf, sizeof (buf), FALSE))
          {
            int                 res = jc_string2reply(&reply, buf);
          }
        }
#endif
      }

    do_action:

      switch (res_action)
      {
        case OPT_REJECT:
          if (done != NULL)
            *done = TRUE;
          log_msg_context(ctx, log_msg);
          (void) jsmfi_setreply(ctx, reply.rcode, reply.xcode, reply.msg);
          result = SMFIS_REJECT;
          break;                /* end of OPT_REJECT */
        case OPT_NOTIFY:
          if (done != NULL)
            *done = TRUE;
          log_msg_context(ctx, log_msg);
          result = do_notify(ctx, ahead, NULL, MSG_XFILE, "XFILE");
          break;
        case OPT_DISCARD:
          if (done != NULL)
            *done = TRUE;
          log_msg_context(ctx, log_msg);
          result = SMFIS_DISCARD;
          break;                /* end of OPT_DISCARD */
        case OPT_X_HEADER:
          if (done != NULL)
            *done = TRUE;
          log_msg_context(ctx, log_msg);
          {
            char               *tag = cf_get_str(CF_XFILE_SUBJECT_TAG);

            if (!add_tag2subject(ctx, tag))
              LOG_MSG_WARNING("Error changing subject");
          }
          result = SMFIS_CONTINUE;
          break;                /* end of OPT_X_HEADER */
        default:
          break;
      }

      (void) smtprate_add_entry(RATE_XFILES, priv->peer_addr,
                                priv->peer_name, 1, now);
    }
    if (nb_exefiles > 0 || (done != NULL && *done))
      log_attached_files(CONNID_STR(priv->id), priv->peer_addr, ahead);
  }

fin:
  return result;
}

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
eom_check_virus(ctx, ahead, done)
     SMFICTX            *ctx;
     attachment_T       *ahead;
     bool               *done;
{
  CTXPRIV_T          *priv = MLFIPRIV(ctx);
  sfsistat            result = SMFIS_CONTINUE;
  char                log_msg_buf[1024];

  if (done != NULL)
    *done = FALSE;
  if (cf_get_int(CF_SCANNER_ACTION) == OPT_OK || !shall_check_virus(ctx))
    return result;

  memset(log_msg_buf, 0, sizeof (log_msg_buf));
  /*
   **
   ** Let's do antivirus checking...
   **
   */
  {
    char                question[JCOMBUFSZ];
    char                why[JCOMBUFSZ];
    char                data[JCOMBUFSZ];
    int                 avres = AV_OK;
    char               *msg = NULL;

    memset(why, 0, sizeof (data));
    memset(data, 0, sizeof (why));
    strlcpy(question, priv->fname, sizeof (question));
    avres = av_client(why, sizeof (why), data, sizeof (data), question);
    MESSAGE_INFO(11, "av_client : %3d (%s) (%s)", avres, why,
                 strchomp(question));

    if (avres == AV_ZERO || avres == AV_OK)
      goto fin;

    if (avres == AV_ERROR)
    {
      if (done != NULL)
        *done = TRUE;
      MESSAGE_ERROR(9, "%s Error with external message scanner",
                    CONNID_STR(priv->id));
      if (cf_get_int(CF_SCANNER_REJECT_ON_ERROR) == OPT_YES)
      {
        (void) jsmfi_vsetreply(ctx, "450", "4.6.0",
                               "Temporary system error. Come back later");
        result = SMFIS_TEMPFAIL;
      }
      log_msg_context(ctx, "External Scanner Error");

      goto fin;
    }

    switch (avres)
    {
      case AV_VIRUS:
        priv->nb_virus++;
        stats_inc(STAT_VIRUS, 1);
        FREE(priv->msg.virus);
        priv->msg.virus = strdup(data);
        log_virus(CONNID_STR(priv->id), priv->peer_addr, data);

        snprintf(log_msg_buf, sizeof (log_msg_buf), "%s : %s", MSG_VIRUS, data);
        msg = MSG_VIRUS;

        priv->fsuffix = SUFFIX_VIRUS;
        if (cf_get_int(CF_SCANNER_SAVE) == OPT_YES)
          DO_QUARANTINE_MESSAGE(priv, WHY_VIRUS, NULL);
        break;
      case AV_POLICY:
        priv->nb_policy++;
        stats_inc(STAT_POLICY, 1);

        snprintf(log_msg_buf, sizeof (log_msg_buf), "%s", MSG_SHORT_POLICY);
        msg = MSG_POLICY;

        priv->fsuffix = SUFFIX_POLICY;
        if (cf_get_int(CF_SCANNER_SAVE) == OPT_YES)
          DO_QUARANTINE_MESSAGE(priv, WHY_POLICY, NULL);
        break;
    }

    switch (cf_get_int(CF_SCANNER_ACTION))
    {
      case OPT_REJECT:
        if (done != NULL)
          *done = TRUE;
        (void) jsmfi_vsetreply(ctx, "550", "5.6.0", "%s : %s", msg, data);
        result = SMFIS_REJECT;
        break;                  /* end of OPT_REJECT */
      case OPT_NOTIFY:
        if (done != NULL)
          *done = TRUE;
        result = do_notify(ctx, ahead, data, msg, "VIRUS");
        break;
      case OPT_DISCARD:
        if (done != NULL)
          *done = TRUE;
        result = SMFIS_DISCARD;
        break;                  /* end of OPT_DISCARD */
      case OPT_X_HEADER:
        /* XXX add x-header */
        if (done != NULL)
          *done = TRUE;
        break;
      default:
        break;
    }

    log_msg_context(ctx, log_msg_buf);
    if (avres == AV_VIRUS)
      log_attached_files(CONNID_STR(priv->id), priv->peer_addr, ahead);
  }

fin:
  return result;
}


/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
static              sfsistat
eom_check_content(ctx, done)
     SMFICTX            *ctx;
     bool               *done;
{
  CTXPRIV_T          *priv = MLFIPRIV(ctx);
  sfsistat            result = SMFIS_CONTINUE;

  /*
   **
   ** Content check (anti-spam)
   */
  priv->netScores.bayes = 0.;
  if (!shall_check_content(ctx))
    return result;

  /*
   ** Bayesian filter
   */
  if (cf_get_int(CF_BAYESIAN_FILTER) == OPT_YES)
  {
    bool                check_it = TRUE;

    priv->rawScores.do_bayes = FALSE;
    priv->rawScores.bayes = 0.;

    check_it = check_policy_all_rcpts("BayesCheck", priv->peer_addr,
                                      priv->peer_name, CTX_NETCLASS_LABEL(priv),
                                      priv->env_from,
                                      priv->env_rcpt, TRUE, OPT_DEFAULT);
    if (check_it)
    {
      sfilter_vsm_T       bcheck;
      double              score = 0;

      memset(&bcheck, 0, sizeof (bcheck));
      bcheck.nbt = cf_get_int(CF_BAYES_NB_TOKENS);
      if (bcheck.nbt <= 0)
        bcheck.nbt = 64;
      score = sfilter_check_message(CONNID_STR(priv->id), priv->fname, &bcheck);
      if (score >= 0.)
        MESSAGE_INFO(10, "%s Bayes filter score : %6.3f",
                     CONNID_STR(priv->id), score);
      else
        MESSAGE_INFO(10, "%s Bayes filter score : Unchecked",
                     CONNID_STR(priv->id));
      if (score >= 0.)
      {
	priv->rawScores.Bayes.value = score;
	priv->rawScores.Bayes.odds  = logit(score);
	priv->rawScores.Bayes.actif = TRUE;
	priv->netScores.Bayes = priv->rawScores.Bayes;

        priv->netScores.bayes = score;
        priv->rawScores.bayes = score;
        priv->rawScores.do_bayes = TRUE;

        if (score >= BSCORE_HI)
          priv->nb_bspam++;
        if (score <= BSCORE_LO)
          priv->nb_bham++;
      } else
      {
	priv->rawScores.Bayes.value = 0.5;
	priv->rawScores.Bayes.odds  = 0.;
	priv->rawScores.Bayes.actif = FALSE;
	priv->netScores.Bayes = priv->rawScores.Bayes;

        priv->netScores.bayes = 0.;
        priv->rawScores.bayes = 0.;
        priv->rawScores.do_bayes = FALSE;
      }
    }
  }
#if _FFR_LOGREG
  if (lr_filter_ok && cf_get_int(CF_BAYESIAN_FILTER) == OPT_YES)
  {
    bool                check_it = TRUE;
    size_t              fsize;
    size_t              fsize_max = 500 KBYTES;

    {
      char                *env;
      
      if ((env = getenv("LR_MAX_FSIZE")) != NULL)
	fsize_max = str2size(env, NULL, fsize_max);
    }
    fsize = get_file_size(priv->fname);

    if (fsize < fsize_max)
      check_it = check_policy_all_rcpts("BayesCheck", priv->peer_addr,
                                      priv->peer_name,
                                      CTX_NETCLASS_LABEL(priv),
                                      priv->env_from,
                                      priv->env_rcpt, TRUE, OPT_DEFAULT);
    else
      check_it = FALSE;

    if (check_it)
    {
      test_score_T        mscore, nscore;

      memset(&mscore, 0, sizeof (mscore));

      if (lr_classify(CONNID_STR(priv->id), priv->fname, &mscore))
        ;

      priv->rawScores.LogReg = priv->netScores.LogReg = mscore;
      {
        char                buf[256];
        double              gres = 0.;
        bool                agree = TRUE;
        int                 win;

	nscore.value = 0.;
	nscore.odds = 0.;
	nscore.actif = FALSE;

	agree = TRUE;
#if 1
	if (priv->rawScores.Bayes.actif && priv->rawScores.LogReg.actif)
	{
	  gres = logitinv(logit(priv->rawScores.Bayes.value) + 
			  logit(priv->rawScores.LogReg.value));
	  agree = (SIGN(priv->rawScores.Bayes.value - 0.5) == 
		   SIGN(priv->rawScores.LogReg.value - 0.5));
	  win = SIGN(fabs(priv->rawScores.Bayes.value - 0.5) - 
		     fabs(priv->rawScores.LogReg.value - 0.5));

	  nscore.value = gres;
	  nscore.odds = logit(gres);
	  nscore.actif = TRUE;
	}
	if (priv->rawScores.Bayes.actif && !priv->rawScores.LogReg.actif)
	{
	  gres = priv->rawScores.Bayes.value;
	  nscore = priv->rawScores.Bayes;
	}
	if (!priv->rawScores.Bayes.actif && priv->rawScores.LogReg.actif)
	{
	  gres = priv->rawScores.LogReg.value;
	  nscore = priv->rawScores.LogReg;
	}

	if (priv->rawScores.Bayes.actif || priv->rawScores.LogReg.actif)
	{
	  priv->rawScores.bayes = gres;
	  priv->rawScores.do_bayes = TRUE;
	  priv->netScores.bayes = gres;
	  priv->netScores.do_bayes = TRUE;
	}

	priv->rawScores.Global = priv->netScores.Global = nscore;
#else
        if (priv->rawScores.bayes > 0)
          gres = logitinv(logit(priv->rawScores.bayes) + logit(mscore.value));
        else
          gres = mscore.value;

        agree = SIGN(priv->rawScores.bayes - 0.5) == SIGN(mscore.value - 0.5);
        win =
          SIGN(fabs(priv->rawScores.bayes - 0.5) - fabs(mscore.value - 0.5));
#endif

        snprintf(buf, sizeof (buf), "%s B=%.5f L=%.5f G=%.5f %-8s %s%s",
                 CONNID_STR(priv->id),
                 priv->rawScores.Bayes.value,
		 priv->rawScores.LogReg.value,
		 gres,
                 STRBOOL(agree, "Agree", "Disagree"),
                 STRBOOL(agree, "", STRBOOL(win > 0, "Winner=B", "Winner=P")),
                 STRBOOL(agree, "", STRBOOL(gres > 0.5, "S", "H")));

        MESSAGE_INFO(10, buf);

        smfi_addheader(ctx, "X-Scores-Stats", buf);
      }
    }
  }
#endif             /* _FFR_LOGREG */

  /* 
   ** Message contents : oracle, URLBL and pattern matching
   */
  {
    bool                do_regex, do_oracle, do_urlbl;

    do_regex = (cf_get_int(CF_SPAM_REGEX) != OPT_NO);
    do_oracle = (cf_get_int(CF_SPAM_ORACLE) != OPT_NO);
    do_urlbl = (cf_get_int(CF_SPAM_URLBL) != OPT_NO);

    if (do_oracle)
      do_oracle = !check_access(priv->peer_addr, priv->netclass.class,
                                priv->peer_name, priv->env_from, NULL,
                                "NO_ORACLE");

    if (do_regex)
      do_regex =
        !check_access(priv->peer_addr, priv->netclass.class, priv->peer_name,
                      priv->env_from, NULL, "NO_PATTERN_CHECK");

    if (do_urlbl)
      do_urlbl =
        !check_access(priv->peer_addr, priv->netclass.class, priv->peer_name,
                      priv->env_from, NULL, "NO_URLBL");

    priv->rawScores.do_regex = do_regex;
    priv->rawScores.do_oracle = do_oracle;
    priv->rawScores.do_urlbl = do_urlbl;

    if (do_regex || do_oracle || do_urlbl)
      result = check_msg_contents(ctx);
  }

  /*
   ** Appel a la routine de decision de score
   */
  {
    bool                do_log = FALSE;

    result = evaluate_message_score(ctx, &do_log);
    if (do_log && log_level >= 8)
    {
      char                logbuf[256];

      snprintf(logbuf, sizeof (logbuf), "%s : B=%.3f U=%d R=%d O=%d -> G=%.3f",
               "Content Check",
               priv->rawScores.bayes, priv->rawScores.urlbl,
               priv->rawScores.body + priv->rawScores.headers,
               priv->rawScores.oracle, priv->rawScores.combined);

      log_msg_context(ctx, logbuf);
    }
  }

  {
    char               *env = getenv("DUMP_MESSAGE_SCORE");

    if (env != NULL && STRCASEEQUAL(env, "yes"))
      dump_msg_scores4stats(ctx);
  }


  (void) smtprate_add_entry(RATE_SCORE, priv->peer_addr, priv->peer_name,
                            (int) (100 * priv->rawScores.combined), time(NULL));
  (void) smtprate_add_entry(RATE_SPAM, priv->peer_addr, priv->peer_name,
                            priv->rawScores.bayes > BSCORE_HI ? 1 : 0,
                            time(NULL));
  (void) smtprate_add_entry(RATE_HAM, priv->peer_addr, priv->peer_name,
                            priv->rawScores.bayes < BSCORE_LO ? 1 : 0,
                            time(NULL));

fin:
  return result;
}
