/*
 *
 * j-chkmail - Mail Server Filter for sendmail
 *
 * Copyright (c) 2001, 2002, 2003, 2004 - Ecole des Mines de Paris
 *
 *  Auteur     : Jose Marcio Martins da Cruz
 *               Jose-Marcio.Martins@ensmp.fr
 *
 *  Historique :
 *  Creation     : janvier 2002
 *
 * This program is free software, but with restricted license :
 *
 * - j-chkmail is distributed only to registered users
 * - j-chkmail license is available only non-commercial applications,
 *   this means, you can use j-chkmail if you make no profit with it.
 * - redistribution of j-chkmail in any way : binary, source in any
 *   media, is forbidden
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *
 * More details about j-chkmail license can be found at j-chkmail
 * web site : http://j-chkmail.ensmp.fr
 */

/* ****************************************************************************
 *                                                                            *
 *                                                                            *
 **************************************************************************** */
sfsistat
mlfi_helo(ctx, helohost)
     SMFICTX            *ctx;
     char               *helohost;
{
  CTXPRIV_T          *priv = MLFIPRIV(ctx);
  int                 result = SMFIS_CONTINUE;

  INIT_CALLBACK_DELAY();

  if (priv == NULL)
  {
    result = SMFIS_TEMPFAIL;
    goto fin;
  }

  INIT_CALLBACK(priv, CALLBACK_EHLO);

  sm_macro_update(ctx, priv->sm);

#if _FFR_DELAYED_REJECT
  if (priv->delayed_result.result != SMFIS_CONTINUE)
  {
    if (priv->callback >= priv->delayed_result.callback)
    {
      (void) jsmfi_setreply(ctx, "550", "5.7.1", priv->delayed_result.reply);
      log_msg_context(ctx, priv->delayed_result.why);
      goto fin;
    }
  }
#endif

  result = check_cpu_load(ctx, CONNID_STR(priv->id), priv->peer_addr,
                          priv->netclass.class);

  if (result != SMFIS_CONTINUE)
  {
    log_msg_context(ctx, "Global load control reject");

    goto fin;
  }

  result = check_filter_open_connections(ctx, CONNID_STR(priv->id),
                                         priv->peer_addr, priv->netclass.class);
  if (result != SMFIS_CONTINUE)
  {
    log_msg_context(ctx, "Global open connections limit");

    goto fin;
  }

  if (helohost == NULL)
  {
    LOG_MSG_WARNING("%s : helohost = NULL", CONNID_STR(priv->id));
    result = SMFIS_TEMPFAIL;

    goto fin;
  }

  FREE(priv->helohost);
  priv->helohost = NULL;

  if ((priv->helohost = strdup(helohost)) == NULL)
  {
    LOG_SYS_ERROR("strdup(helohost) error");
    result = SMFIS_TEMPFAIL;

    goto fin;
  }

  priv->ehlo_flags = check_ehlo_conformance(ctx, log_level >= 10);

  result = validate_connection(ctx);
  if (result != SMFIS_CONTINUE)
    goto fin;

#if _FFR_MODULES
  /* 
   ** j-chkmail modules
   **
   */
  if (do_module_callback(ctx, 0, &result))
    goto fin;
  if (result != SMFIS_CONTINUE)
    goto fin;
#endif             /* _FFR_MODULES */

#if 0
  /* check HELO contents */
  if ((helohost != NULL) && (cf_get_int(CF_SPAM_REGEX) == OPT_YES))
  {
    LOG_MSG_DEBUG(15, "check_helo_content");

    if (shall_check_content(ctx))
    {
      int                 score_min = cf_get_int(CF_REGEX_MAX_SCORE);
      int                 score = 0;

      score =
        check_regex(CONNID_STR(priv->id), priv->peer_addr, helohost, MAIL_HELO);
      priv->rawScores.headers += score;
      if (score >= score_min)
      {
        result = SMFIS_REJECT;

        stats_inc(STAT_HELO_CONTENTS, 1);
#if _FFR_DBPOLICY
        (void) get_reply_msg(ctx, "MSG_HELO_CONTENTS", reply, sizeof (reply),
                             NULL, NULL);
        (void) jsmfi_setreply(ctx, "550", "5.7.1", reply);
#else
        (void) jsmfi_setreply(ctx, "550", "5.7.1", MSG_HELO_CONTENTS);
#endif
        log_msg_context(ctx, MSG_HELO_CONTENTS);

        goto fin;
      }
    }
  }
#endif

fin:
  CHECK_CALLBACK_DELAY();

  /* continue processing */
  return result;
}
